//ProjTess par Marc Plouhinec
//email: m_plouhinec@yahoo.fr
//site: www.thebestof.tk

//Ce programme sert d'exemple pour mon TIPE "La tetra-espace"
//Je me suis inspir de la mthode dcrite par Steve R. Hollasch (http://research.microsoft.com/~hollasch/thesis/)

//Date de cration de ce fichier: 18/06/2004; 16:25:57
//Dernire mise  jour: 19/06/2004; 13:10:21

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files


//Constantes:
#define Pi 3.1416
#define Vangle4 Pi/2
#define Vangle3 Pi/2
#define DecalageX 75
#define DecalageY 55

#define ESC_KEY 264
#define KEY_PLUS 43
#define KEY_MINUS 45
#define KEY_MUL 42
#define KEY_DIV 47
#define KEY_0 48
#define KEY_1 49
#define KEY_2 50
#define KEY_3 51
#define KEY_4 52
#define KEY_5 53
#define KEY_6 54
#define KEY_7 55
#define KEY_8 56
#define KEY_9 57

//On dfinit les coordonnes des vecteurs de l'oeil
float from3[] = {3,0.99,1.82};
float from4[] = {4,0,0,0};
float over4[] = {0,0,1,0};
float to3[] = {0,0,0};
float to4[] = {0,0,0,0};
float up3[] = {0,1,0};
float up4[] = {0,1,0,0};

int rp[2][6] = {{0,1,2,0,1,2},{1,2,0,3,3,3}};

//On dfinit les coordonnes des points des sommets du tesseract
char TesseractPoints[16][4] = {{-1,-1,-1,-1},{-1,-1,-1,1},{-1,-1,1,1},{-1,-1,1,-1},{-1,1,-1,-1},{-1,1,-1,1},{-1,1,1,1},{-1,1,1,-1},{1,-1,-1,-1},{1,-1,-1,1},{1,-1,1,1},{1,-1,1,-1},{1,1,-1,-1},{1,1,-1,1},{1,1,1,1},{1,1,1,-1}};
//On dfinit l'ordre dans lequel on va lier les points par des lignes:
char TesseractOrdre[47] = {1,2,3,4,1,0,2,6,14,10,2,0,1,5,13,9,1,0,3,7,15,11,3,0,4,8,16,12,4,0,9,10,11,12,9,0,5,6,7,8,5,0,13,14,15,16,13};

//Variables:
float vproj3[16][3];
float vx[16];
float vy[16];
float vdepth[16];

float va[3];
float vb[3];
float vc[3];

float wa[4];
float wb[4];
float wc[4];
float wd[4];

float pconst;
float NormeDeFrom3;
float NormeDeFrom4;
int Zoom = 350;
char ModeDeProjection = 3;
//0 4D->3D parrallle  3D->2D parrallle
//1 4D->3D parrallle  3D->2D perspective
//2 4D->3D perspective  3D->2D parrallle
//3 4D->3D perspective  3D->2D perspective
float Data4Radius;

////////////////////////////////////////////////////////////////////////////////////////////////////////
//Fonctions de bases

// Oprations sur les vecteurs en 3D

//va = vb-vc
void SoustractionDeVecteurs3D( float *va, float *vb,float *vc )
{
	int i;
	for (i=0; i<3; i++)
	{
		va[i] = vb[i] - vc[i];
	}
}

float NormeVecteur3D( float *v )
{
	return sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
}

//va = vb/s
void DiviseVecteurParUnScalaire3D( float *va , float *vb, float s)
{
	int i;
	for (i=0; i<3; i++)
	{
		va[i] = vb[i]/s;
	}
}

//va = vb*s
void MultiplieVecteurParUnScalaire3D( float *va , float *vb, float s)
{
	int i;
	for (i=0; i<3; i++)
	{
		va[i] = vb[i]*s;
	}
}

//va = X3(vb,vc) avec X3(u,v) le produit vectoriel de u et v
void ProduitVectoriel3D( float *va, float *vb, float *vc)
{
	va[0] = (vb[1] * vc[2]) - (vb[2] * vc[1]);
	va[1] = (vb[2] * vc[0]) - (vb[0] * vc[2]);
	va[2] = (vb[0] * vc[1]) - (vb[1] * vc[0]);
}

//v = TesseractPoints[t]
void StockeVproj33D(float *v, int t)
{
	int i;
	for (i=0; i<3; i++)
	{
		v[i] = vproj3[t][i];
	}
}

//va.vb
float ProduitScalaire3D( float *va, float *vb)
{
	return vb[0]*va[0]+vb[1]*va[1]+vb[2]*va[2];
}


// Oprations sur les vecteurs en 4D

//va = vb-vc
void SoustractionDeVecteurs4D( float *va, float *vb,float *vc )
{
	int i;
	for (i=0; i<4; i++)
	{
		va[i] = vb[i] - vc[i];
	}
}

float NormeVecteur4D( float *v )
{
	return sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]+v[3]*v[3]);
}

//va = vb/s
void DiviseVecteurParUnScalaire4D( float *va , float *vb, float s)
{
	int i;
	for (i=0; i<4; i++)
	{
		va[i] = vb[i]/s;
	}
}

//va = vb*s
void MultiplieVecteurParUnScalaire4D( float *va , float *vb, float s)
{
	int i;
	for (i=0; i<4; i++)
	{
		va[i] = vb[i]*s;
	}
}

//va = X4(vb,vc,vd) avec X3(u,v,w) le produit vectoriel de u, v et w
void  ProduitVectoriel4D(float *va, float *vb, float *vc, float *vd)
{
   float A, B, C, D, E, F;

   A = (vc[0] * vd[1]) - (vc[1] * vd[0]);
   B = (vc[0] * vd[2]) - (vc[2] * vd[0]);
   C = (vc[0] * vd[3]) - (vc[3] * vd[0]);
   D = (vc[1] * vd[2]) - (vc[2] * vd[1]);
   E = (vc[1] * vd[3]) - (vc[3] * vd[1]);
   F = (vc[2] * vd[3]) - (vc[3] * vd[2]);

   va[0] =   (vb[1] * F) - (vb[2] * E) + (vb[3] * D);
   va[1] = - (vb[0] * F) + (vb[2] * C) - (vb[3] * B);
   va[2] =   (vb[0] * E) - (vb[1] * C) + (vb[3] * A);
   va[3] = - (vb[0] * D) + (vb[1] * B) - (vb[2] * A);
}

//v = TesseractPoints[t]
void StockeTesseractPoints4D(float *v, int t)
{
	int i;
	for (i=0; i<4; i++)
	{
		v[i] = TesseractPoints[t][i];
	}
}

//va.vb
float ProduitScalaire4D( float *va, float *vb)
{
	return vb[0]*va[0]+vb[1]*va[1]+vb[2]*va[2]+vb[3]*va[3];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//Fonctions avances
void Presentation()
{
	clrscr ();
	printf("         ProjTess\n    par Marc Plouhinec\n     www.thebestof.tk\n   m_plouhinec@yahoo.fr");
	ngetchx ();
	clrscr ();
	printf("Touches:\nRotations:\n1-XY\n2-YZ\n3-ZX\n4-XW\n5-YW\n6-ZW\n");
	ngetchx();
	clrscr();
	printf("Autre:\nESC-Quitter\n0-reset\n+-Zoom+\n--Zoom-\n*et/-Type de projection");
  ngetchx ();
}

void Init()
{
	Data4Radius = 0;
	int ii;
	int dist;
	float TempV[4],TempV2[4];
  for (ii=0;  ii < 16;  ii++)
  {
  	StockeTesseractPoints4D(TempV2, ii);
  	SoustractionDeVecteurs4D(TempV, TempV2, to4);
		dist = ProduitScalaire4D(TempV,TempV);
		if (dist > Data4Radius)
		Data4Radius = dist;
	}
	Data4Radius = sqrt (Data4Radius);
}

void CalcV3()
{
	//Calcule vc
	SoustractionDeVecteurs3D(vc,to3,from3);
	DiviseVecteurParUnScalaire3D( vc,vc,NormeVecteur3D(vc));
	//Calcule va
	ProduitVectoriel3D(va, vc, up3);
	DiviseVecteurParUnScalaire3D( va,va,NormeVecteur3D(va));
	//Calcule vb
	ProduitVectoriel3D(vb, va, vc);
	
/*	clrscr();
	printf("va={%f,%f,%f}\nvb={%f,%f,%f}\nvc={%f,%f,%f}\n",va[0],va[1],va[2],vb[0],vb[1],vb[2],vc[0],vc[1],vc[2]);
	ngetchx ();*/
}

void CalcV4()
{
	//Calcule wd
	SoustractionDeVecteurs4D(wd,to4,from4);
	DiviseVecteurParUnScalaire4D( wd,wd,NormeVecteur4D(wd));
	//Calcule wa
	ProduitVectoriel4D(wa, up4, over4, wd);
	DiviseVecteurParUnScalaire4D( wa,wa,NormeVecteur4D(wd));
	//Calcule wb
	ProduitVectoriel4D(wb, over4, wd, wa);
	DiviseVecteurParUnScalaire4D( wb,wb,NormeVecteur4D(wb));
	//Calcule wc
	ProduitVectoriel4D(wc, wd, wa, wb);
	
	/*clrscr();
	printf("wa={%f,%f,%f,%f}\nwb={%f,%f,%f,%f}\nwc={%f,%f,%f,%f}\nwd={%f,%f,%f,%f}\n",wa[0],wa[1],wa[2],wa[3],wb[0],wb[1],wb[2],wb[3],wc[0],wc[1],wc[2],wc[3],wd[0],wd[1],wd[2],wd[3]);
	ngetchx ();*/
}

void ProjectTo3D()
{
	float TempV[4],TempV2[4],rtemp=0;
	if(ModeDeProjection >= 2) //perspective
		{
			pconst = 1/(tan(Vangle4/2));
		}
		else
		{
			rtemp = 1/Data4Radius;
		}
	int ii = 0;
	while(ii<16)
	{
		StockeTesseractPoints4D(TempV2, ii);
		SoustractionDeVecteurs4D(TempV,TempV2,from4);
		
		vdepth[ii]=ProduitScalaire4D(wd,TempV);
		
		if(ModeDeProjection >= 2) //perspective
		{
			rtemp=pconst/(vdepth[ii]);
		}
		
		
		vproj3[ii][0]=(ProduitScalaire4D(wa,TempV))*rtemp;
		vproj3[ii][1]=(ProduitScalaire4D(wb,TempV))*rtemp;
		vproj3[ii][2]=(ProduitScalaire4D(wc,TempV))*rtemp;
		
		ii++;
	}
	
	/*for (ii =0 ; ii<16; ii++)
	{
		clrscr();
		printf("vproj3[%d]={%f,%f,%f}\n",ii,vproj3[ii][0],vproj3[ii][1],vproj3[ii][2]);
		ngetchx ();
	}*/
	
}

void ProjectToScreen()
{
	float TempV[4],TempV2[4],rtemp=0;
	if(ModeDeProjection == 1 || ModeDeProjection==3)//perspective
	{
		pconst = 1/(tan(Vangle3/2));
	}
	else
	{
		rtemp = 1/ sqrt(3);
	}
	int ii = 0;
	while(ii<16)
	{
		StockeVproj33D(TempV2, ii);
		SoustractionDeVecteurs3D(TempV,TempV2,from3);
		
		if(ModeDeProjection == 1 || ModeDeProjection==3)//perspective
		{
			rtemp=pconst/(ProduitScalaire3D(vc,TempV));
		}
		
		vx[ii]=rtemp*ProduitScalaire3D(va,TempV);
		vy[ii]=rtemp*ProduitScalaire3D(vb,TempV);
		
		ii++;
	}
	
	/*for (ii =0 ; ii<16; ii++)
	{
		clrscr();
		printf("vx[%d]=%f\n",ii,vx[ii]);
		printf("vy[%d]=%f\n",ii,vy[ii]);
		ngetchx ();
	}*/
}

void Draw()
{
	clrscr();
	int t;
	for (t = 1; t< 47; t++)
	{
		if (TesseractOrdre[t]==0)
		{
			t=t+2;
		}
		DrawLine(vx[TesseractOrdre[t-1]-1]*Zoom+DecalageX,vy[TesseractOrdre[t-1]-1]*Zoom+DecalageY,vx[TesseractOrdre[t]-1]*Zoom+DecalageX,vy[TesseractOrdre[t]-1]*Zoom+DecalageY,A_NORMAL);
	}
}

void Rot3DView(float cos2, float sin2, int rp1, int rp2)
{
	from3[rp1]=cos2*(from3[rp1]-to3[rp1])+sin2*(from3[rp2]-to3[rp2])+to3[rp1];
	from3[rp2]=cos2*(from3[rp2]-to3[rp2])-sin2*(from3[rp1]-to3[rp1])+to3[rp2];
	DiviseVecteurParUnScalaire3D( from3,from3,NormeVecteur3D(from3));//Palie aux bugs d aux approximations
	MultiplieVecteurParUnScalaire3D(from3,from3,NormeDeFrom3);       //

	up3[rp1]=cos2*up3[rp1]+sin2*up3[rp2];
	up3[rp2]=cos2*up3[rp2]-sin2*up3[rp1];
	DiviseVecteurParUnScalaire3D( up3,up3,NormeVecteur3D(up3));//Palie aux bugs d aux approximations

	CalcV3();
	ProjectToScreen();
}

void Rot4DView(float cos2, float sin2, int rp1, int rp2)
{
	from4[rp1]=cos2*(from4[rp1]-to4[rp1])+sin2*(from4[rp2]-to4[rp2])+to4[rp1];
	from4[rp2]=cos2*(from4[rp2]-to4[rp2])-sin2*(from4[rp1]-to4[rp1])+to4[rp2];
	DiviseVecteurParUnScalaire4D( from4,from4,NormeVecteur4D(from4));//Palie aux bugs d aux approximations
	MultiplieVecteurParUnScalaire4D(from4,from4,NormeDeFrom4);       //
	
	up4[rp1]=cos2*up4[rp1]+sin2*up4[rp2];
	up4[rp2]=cos2*up4[rp2]-sin2*up4[rp1];
	DiviseVecteurParUnScalaire4D( up4,up4,NormeVecteur4D(up4));//Palie aux bugs d aux approximations
	
	over4[rp1]=cos2*over4[rp1]+sin2*over4[rp2];
	over4[rp2]=cos2*over4[rp2]-sin2*over4[rp1];
	DiviseVecteurParUnScalaire4D( over4,over4,NormeVecteur4D(over4));//Palie aux bugs d aux approximations
	
	CalcV4();
	ProjectTo3D();
	ProjectToScreen();
}

void Rotations()
{
	int key=0;
	float delta;
	NormeDeFrom3 = NormeVecteur3D(from3);
	NormeDeFrom4 = NormeVecteur4D(from4);
	while(key!=ESC_KEY)
	{
		key = ngetchx();
		//Rotations
		if(key == KEY_1)
		{
			delta=(100*Pi)/(4*360);
			Rot3DView(cos(delta),sin(delta),rp[0][0],rp[1][0]);
		}
		if(key == KEY_2)
		{
			delta=(100*Pi)/(4*360);
			Rot3DView(cos(delta),sin(delta),rp[0][1],rp[1][1]);
		}
		if(key == KEY_3)
		{
			delta=(100*Pi)/(4*360);
			Rot3DView(cos(delta),sin(delta),rp[0][2],rp[1][2]);
		}
		if(key == KEY_4)
		{
			delta=(100*Pi)/(4*360);
			Rot4DView(cos(delta),sin(delta),rp[0][3],rp[1][3]);
		}
		if(key == KEY_5)
		{
			delta=(100*Pi)/(4*360);
			Rot4DView(cos(delta),sin(delta),rp[0][4],rp[1][4]);
		}
		if(key == KEY_6)
		{
			delta=(100*Pi)/(4*360);
			Rot4DView(cos(delta),sin(delta),rp[0][5],rp[1][5]);
		}
		//Actions diverses
		if(key == KEY_PLUS)
		{
			Zoom=Zoom+10;
		}
		if(key == KEY_MINUS)
		{
			Zoom=Zoom-10;
		}
		if(key == KEY_MUL && ModeDeProjection!=3)
		{
			ModeDeProjection++;
			ProjectTo3D();
			ProjectToScreen();
		}
		if(key == KEY_DIV && ModeDeProjection!=0)
		{
			ModeDeProjection--;
			ProjectTo3D();
			ProjectToScreen();
		}
		if(key == KEY_0)//reset
		{
			from3[0]=3;
			from3[1]=0.99;
			from3[2]=1.82;

			int i,j;
			for(i=0; i<16; i++)
				for(j=0; j<3; j++)
				{
					vproj3[i][j]=0;
					vx[i]=0;
					vy[i]=0;
					vdepth[i]=0;
				}
			
			for(i=0; i<3; i++)
			{
				va[i]=0;
				vb[i]=0;
				vc[i]=0;
				to3[i]=0;
				up3[i]=0;
			}

			for(i=0; i<4; i++)
			{
				wa[i]=0;
				wb[i]=0;
				wc[i]=0;
				wd[i]=0;
				from4[i]=0;
				over4[i]=0;
				to4[i]=0;
				up4[i]=0;
			}
			from4[0]=4;
			over4[2]=1;
			up3[1]=1;
			up4[1]=1;
			
			Zoom = 350;
			ModeDeProjection = 3;

			Init();
			CalcV3();
			CalcV4();
			ProjectTo3D();
			ProjectToScreen();
		}
		Draw();
		if(ModeDeProjection == 0)
			printf("3D //  4D //\n");
		if(ModeDeProjection == 1)
			printf("3D #   4D //\n");
		if(ModeDeProjection == 2)
			printf("3D //  4D # \n");
		if(ModeDeProjection == 3)
			printf("3D #   4D # \n");
	}

}

// Main Function
void _main(void)
{
	Presentation();
	Init();
	CalcV3();
	CalcV4();
	ProjectTo3D();
	ProjectToScreen();
	Draw();
	Rotations();
}
